package algorithms.graphColorer.heuristics;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import org.jgrapht.UndirectedGraph;
import org.jgrapht.graph.MaskFunctor;

import algorithms.graphColorer.Colorer;
import dataStructures.graph.Edge;
import dataStructures.graph.ExtendedUndirectedGraph;
import dataStructures.graph.GraphColoring;

public class GreedyColorer<V, E> implements Colorer<V, E>
{

	class SimpleRevertMaskFunction implements MaskFunctor<V, E>
	{
		private final HashSet<V> maskedVertices;

		public SimpleRevertMaskFunction()
		{
			this.maskedVertices = new HashSet<V>();
		}

		public void addVertex(V vertex)
		{
			this.maskedVertices.add(vertex);
		}

		public boolean isEdgeMasked(E edge)
		{
			return false;
		}

		public boolean isVertexMasked(V vertex)
		{
			return !this.maskedVertices.contains(vertex);
		}

	}

	public static int maxC = 8;

	public static int pow = 10;

	public static <V, E> int getBestColorForVertex(UndirectedGraph<V, E> graph, GraphColoring<V, E> coloring, V vertex,
			int colorStrategy)
	{
		final ArrayList<Integer> possibleColors = new ArrayList<Integer>();
		final HashMap<Integer, HashSet<V>> colorVertexMap = coloring.getColorVertexMapping();
		final Set<Integer> colorSet = colorVertexMap.keySet();
		for (final int color : colorSet)
		{
			final HashSet<V> colorClass = colorVertexMap.get(color);
			boolean flag = true;
			for (final V v : colorClass)
			{
				if ( graph.containsEdge(vertex, v) )
				{
					flag = false;
					break;
				}
			}
			if ( flag )
			{
				possibleColors.add(color);
			}
		}

		Collections.sort(possibleColors);
		if ( possibleColors.size() > 1 )
		{
			if ( colorStrategy == GreedyStrategy.SIMPLE_GREEDY_STRATEGY )
				return possibleColors.get(0);
			else if ( colorStrategy == GreedyStrategy.REVERSE_ORDER_GREEDY_STRATEGY )
				return possibleColors.get(possibleColors.size() - 1);
			else if ( colorStrategy == GreedyStrategy.RANDOM_SEQUENCE_GREEDY_STRATEGY )
				return possibleColors.get(new Random().nextInt(possibleColors.size()));
			else if ( colorStrategy == GreedyStrategy.LARGEST_FIRST_GREEDY_STRATEGY )
			{
				final HashMap<Integer, HashSet<V>> colorClasses = coloring.getColorVertexMapping();

				int score = Integer.MIN_VALUE;
				int color = -1;

				for (final int possibleColor : possibleColors)
				{
					final int size = colorClasses.get(possibleColor).size();
					if ( size > score )
					{
						score = size;
						color = possibleColor;
					}
				}
				return color;
			}
			else if ( colorStrategy == GreedyStrategy.SMALLEST_FIRST_GREEDY_STRATEGY )
			{
				final HashMap<Integer, HashSet<V>> colorClasses = coloring.getColorVertexMapping();

				int score = Integer.MAX_VALUE;
				int color = -1;

				for (final int possibleColor : possibleColors)
				{
					final int size = colorClasses.get(possibleColor).size();
					if ( size < score )
					{
						score = size;
						color = possibleColor;
					}
				}
				return color;
			}
		}
		if ( !possibleColors.isEmpty() )
			return possibleColors.get(0);
		return colorSet.size() == 0 ? 1 : Collections.max(colorSet) + 1;
	}

	public static GraphColoring<Integer, Edge<Integer>> getBestGreedyColoring(
			final ExtendedUndirectedGraph<Integer, Edge<Integer>> graph)
	{
		GraphColoring<Integer, Edge<Integer>> greedyColoring = null;

		{
			int greedyColorNumber = Integer.MAX_VALUE;
			for (final int colorStrategy : GreedyStrategy.COLOR_STRATEGY)
			{
				final GraphColoring<Integer, Edge<Integer>> coloring = new GreedyColorer<Integer, Edge<Integer>>(graph,
						new GreedyStrategy<Integer, Edge<Integer>>(colorStrategy)).getColoring();
				final int numberOfColors = coloring.getNumberOfColors();
				if ( numberOfColors < greedyColorNumber )
				{
					greedyColorNumber = numberOfColors;
					greedyColoring = coloring;
				}
			}
		}
		return greedyColoring;
	}

	public static int getMaxC(int maxColor)
	{
		if ( maxColor <= 2 )
			return GreedyColorer.maxC;
		final double m = (GreedyColorer.pow * Math.log(10) + Math.log(1d / maxColor)) / Math.log(maxColor - 1);
		return Math.min((int) Math.round(m), GreedyColorer.maxC);
	}

	private GraphColoring<V, E> coloring;

	private final UndirectedGraph<V, E> g;

	final GreedyStrategy<V, E> strategy;

	public GreedyColorer(UndirectedGraph<V, E> g, GreedyStrategy<V, E> greedyStrategy)
	{
		this.g = g;
		this.strategy = greedyStrategy;
	}

	private GraphColoring<V, E> color()
	{
		final GraphColoring<V, E> color = new GraphColoring<V, E>(this.g);

		final Collection<V> sortedVertexSet = this.strategy.getSortedVertexSet(this.g);
		for (final V vertex : sortedVertexSet)
		{
			color.setColor(vertex, GreedyColorer.getBestColorForVertex(this.g, color, vertex, this.strategy.getColorStrategy()));
		}
		return color;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see algorithms.Colorer#getColoring()
	 */
	public GraphColoring<V, E> getColoring()
	{
		if ( this.coloring == null )
		{
			this.coloring = this.color();
		}
		return this.coloring;
	}

}