package algorithms.graphColorer.heuristics;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;

import algorithms.graphColorer.Colorer;
import algorithms.graphColorer.reoptimizer.IncConsColoringSolver;
import dataStructures.graph.ExtendedUndirectedGraph;
import dataStructures.graph.GraphColoring;
import dataStructures.utils.ArrayS;

public class SearchTreeColorer<V, E> implements Colorer<V, E>
{
	public static final boolean verbose = false;

	public final HashSet<V> alreadyModifiedVertices = new HashSet<V>();

	private GraphColoring<V, E> coloring;

	private final ExtendedUndirectedGraph<V, E> g;

	private final GreedyStrategy<V, E> strategy = new GreedyStrategy<V, E>(GreedyStrategy.LARGEST_FIRST_GREEDY_STRATEGY);

	public SearchTreeColorer(ExtendedUndirectedGraph<V, E> g)
	{
		this.g = g;
	}

	private int applySearchTreeAlgorithm(final HashSet<V> maskedVertices, final V vertex, final int c, int newColor)
	{
		final TreeMap<Integer, Vector<V>> neighborsPerColorClass = IncConsColoringSolver.getNeighborsForVertexPerColor(this.g,
				vertex, maskedVertices, this.coloring);
		// this.minNeighbours() <= c
		this.alreadyModifiedVertices.add(vertex);
		for (final int i : this.getBestBranchOrder(neighborsPerColorClass))
		{
			this.coloring.setColor(vertex, i);
			final Vector<V> conflicts = new Vector<V>(neighborsPerColorClass.get(i));

			if ( this.solve(this.g, maskedVertices, conflicts, c) )
			{
				newColor = this.coloring.getColor(vertex);
				break;
			}
		}
		this.alreadyModifiedVertices.remove(vertex);
		return newColor;
	}

	private GraphColoring<V, E> color()
	{
		this.coloring = new GraphColoring<V, E>(this.g);

		final HashSet<V> maskedVertices = new HashSet<V>();
		maskedVertices.addAll(this.g.vertexSet());

		final ArrayList<V> sortedVertexSet = this.strategy.getSortedVertexSet(this.g);
		int noDeniedVertices = 0;
		int maxColor = 1;

		int step = 1;
		final int maxStep = sortedVertexSet.size() / 10;
		int stepsWithoutSuccess = 0;
		int backindex = -1;

		for (int index = 0; index < sortedVertexSet.size(); index++)
		{
			final V vertex = sortedVertexSet.get(index);
			maskedVertices.remove(vertex);

			// greedy color
			int newColor = GreedyColorer.getBestColorForVertex(this.g, this.coloring, vertex, this.strategy.getColorStrategy());

			// apply search tree algorithm ??
			if ( newColor > maxColor && ++noDeniedVertices == step )
			{
				final int c = this.getC(maxColor, stepsWithoutSuccess, backindex);
				noDeniedVertices = 0;

				if ( SearchTreeColorer.verbose )
				{
					System.out.print("#color:" + this.coloring.getNumberOfColors() + "\tindex:" + index + "\tstep:" + step
							+ "\tc=" + c + "\tback:" + backindex);
				}

				final long begin = System.currentTimeMillis();

				// search tree algorithm
				newColor = this.applySearchTreeAlgorithm(maskedVertices, vertex, c, newColor);

				if ( newColor > maxColor )
				{// fails
					stepsWithoutSuccess++;
					step = this.getStep(maxStep, stepsWithoutSuccess, backindex, index);
				}
				else
				{// success
					if ( stepsWithoutSuccess > 0 && backindex >= 0 )
					{
						if ( SearchTreeColorer.verbose )
						{
							System.out.println("\ngo back to " + backindex + " from " + index);
						}

						for (int t = backindex; t <= index; t++)
						{
							maskedVertices.add(sortedVertexSet.get(t));
							this.coloring.removeColor(sortedVertexSet.get(t));
						}
						maxColor = Collections.max(this.coloring.getColorVertexMapping().keySet());
						index = backindex - 1;
						newColor = -1;
						step = this.getStep(maxStep, stepsWithoutSuccess - 1, backindex, index);
						noDeniedVertices = step - 1;
					}
					else
					{
						step = 1;
					}
					stepsWithoutSuccess = 0;
				}

				backindex = -1;

				if ( SearchTreeColorer.verbose )
				{
					System.out.println("\t" + (newColor <= maxColor) + "\t" + (System.currentTimeMillis() - begin) / 1000d);
				}
			}
			else if ( newColor > maxColor && noDeniedVertices <= Math.max((step / 2), 1) )
			{// save first denied vertex
				backindex = index;
			}

			final boolean flag = newColor > maxColor;
			if ( newColor > 0 )
			{
				this.coloring.setColor(vertex, newColor);
				maxColor = Math.max(maxColor, newColor);
			}

			if ( flag )
			{
				// greedyColor();
			}
		}
		return this.coloring;
	}

	private boolean containsNoModifiedElement(Vector<V> colorClass)
	{
		for (final V v : colorClass)
		{
			if ( this.alreadyModifiedVertices.contains(v) )
				return false;
		}
		return true;
	}

	private int[] getBestBranchOrder(TreeMap<Integer, Vector<V>> colorClass)
	{
		final Set<Integer> colorSet = this.coloring.getColorVertexMapping().keySet();
		final Iterator<Integer> colorIter = colorSet.iterator();

		final int[] colors = new int[colorSet.size()];
		final int[] size = new int[colors.length];

		for (int i = 0; i < colors.length; i++)
		{
			final int color = colorIter.next();
			size[i] = colorClass.containsKey(color) ? colorClass.get(color).size() : Integer.MIN_VALUE;
			colors[i] = color;
		}

		final int[] indicePermutation = ArrayS.sort(size);
		final int[] sortedColors = new int[colors.length];
		for (int i = 0; i < colors.length; i++)
		{
			sortedColors[i] = colors[indicePermutation[i]];
		}
		return sortedColors;
	}

	private int getC(int maxColor, int stepsWithoutSuccess, int backindex)
	{
		int c = Math.max(1, GreedyColorer.getMaxC(maxColor));

		if ( stepsWithoutSuccess > 0 )
		{
			c--;
		}
		if ( stepsWithoutSuccess > 1 )
		{
			c--;
		}
		if ( stepsWithoutSuccess > 2 )
		{
			c--;
		}
		c = Math.max(stepsWithoutSuccess <= 3 ? 3 : 2, c);
		return c;
	}

	public GraphColoring<V, E> getColoring()
	{
		if ( this.coloring == null )
		{
			this.coloring = this.color();
		}
		return this.coloring;
	}

	private int getStep(final int maxStep, int stepsWithoutSuccess, int backindex, int index)
	{
		return Math.max(backindex == index ? 2 : 1, Math.min(maxStep, stepsWithoutSuccess * stepsWithoutSuccess));
	}

	private void greedyColor()
	{
		final GraphColoring<V, E> color = new GraphColoring<V, E>(this.g);
		final HashMap<Integer, HashSet<V>> colorVertexMap = this.coloring.getColorVertexMapping();
		for (final int colorV : ArrayS.randomPermutation(colorVertexMap.keySet()))
		{
			final HashSet<V> set = colorVertexMap.get(colorV);
			for (final V vColor : set)
			{
				final Integer[] neighborColors = this.g.getNeighborsPerColor(vColor, color).keySet().toArray(new Integer[0]);
				Arrays.sort(neighborColors);
				// first
				int newc = neighborColors.length > 0 ? neighborColors[neighborColors.length - 1] + 1 : 1;
				for (int i = 0; i < neighborColors.length - 1; i++)
				{
					if ( neighborColors[i + 1] - neighborColors[i] > 1 )
					{
						newc = neighborColors[i] + 1;
						break;
					}
				}
				color.setColor(vColor, newc);
			}
		}
		this.coloring = color;
	}

	private int minNeighbours(TreeMap<Integer, Vector<V>> colorClass)
	{
		int min = Integer.MAX_VALUE;

		final Collection<Vector<V>> values = colorClass.values();

		for (final Vector<V> vector : values)
		{
			min = Math.min(min, vector.size());
		}
		return min;

	}

	private boolean solve(ExtendedUndirectedGraph<V, E> graph, HashSet<V> maskedVertices, Vector<V> conflicts, int param)
	{
		if ( conflicts.size() == 0 )
			return true;
		else if ( conflicts.size() > param )
			return false;

		final V firstElement = conflicts.firstElement();
		this.alreadyModifiedVertices.add(firstElement);

		final TreeMap<Integer, Vector<V>> colorClass = IncConsColoringSolver.getNeighborsForVertexPerColor(graph, firstElement,
				maskedVertices, this.coloring);
		final int color = this.coloring.getColor(firstElement);

		for (final int i : this.getBestBranchOrder(colorClass))
		{
			if ( i != color && (!colorClass.containsKey(i) || this.containsNoModifiedElement(colorClass.get(i))) )
			{
				this.coloring.setColor(firstElement, i);
				final Vector<V> conflictTmp = new Vector<V>();
				conflictTmp.addAll(conflicts);
				conflictTmp.remove(firstElement);
				if ( colorClass.containsKey(i) )
				{
					conflictTmp.addAll(colorClass.get(i));
				}

				if ( this.solve(graph, maskedVertices, conflictTmp, param - 1) )
					return true;
			}
		}
		this.alreadyModifiedVertices.remove(firstElement);
		this.coloring.setColor(firstElement, color);
		return false;
	}
}