package algorithms.graphColorer.reoptimizer;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import org.jgrapht.UndirectedGraph;

import algorithms.graphColorer.heuristics.GreedyColorer;
import algorithms.graphColorer.heuristics.GreedyStrategy;
import dataStructures.graph.Edge;
import dataStructures.graph.ExtendedSimpleGraphAjazenzList;
import dataStructures.graph.ExtendedUndirectedGraph;
import dataStructures.graph.GraphColoring;
import dataStructures.graph.VCounter;
import dataStructures.utils.ArrayS;

public class IncConsColoringSolver<V, E> extends ColoringReoptimizer<V, E>
{
	public static int maxIteration = 3;

	public static <V, E> TreeMap<Integer, Vector<V>> getNeighborsForVertexPerColor(ExtendedUndirectedGraph<V, E> graph, V vertex,
			HashSet<V> maskedVertices, GraphColoring<V, E> coloring)
	{
		final Collection<V> neighbors = graph.getNeighbors(vertex);
		final TreeMap<Integer, Vector<V>> out = new TreeMap<Integer, Vector<V>>();
		for (final V n : neighbors)
		{
			if ( !maskedVertices.contains(n) )
			{
				final int color = coloring.getColor(n);
				final Vector<V> colorClass;
				if ( out.containsKey(color) )
				{
					colorClass = out.get(color);
				}
				else
				{
					colorClass = new Vector<V>(color);
					out.put(color, colorClass);
				}
				colorClass.add(n);
			}
		}

		return out;
	}

	public final HashSet<V> alreadyModifiedVertices;

	public int iteration = 0;

	public IncConsColoringSolver(ExtendedUndirectedGraph<V, E> graph, GraphColoring<V, E> iniColoring)
	{
		super(graph, iniColoring);
		this.alreadyModifiedVertices = new HashSet<V>();
	}

	private boolean containsNoModifiedElement(Vector<V> colorClass)
	{
		for (final V v : colorClass)
		{
			if ( this.alreadyModifiedVertices.contains(v) )
				return false;
		}
		return true;
	}

	private int[] getBestBranchOrder(final int chosenColor, TreeMap<Integer, Vector<V>> colorClass)
	{
		final Set<Integer> colorSet = this.coloring.getColorVertexMapping().keySet();
		final Iterator<Integer> colorIter = colorSet.iterator();
		final int[] colors = colorSet.contains(chosenColor) ? new int[colorSet.size() - 1] : new int[colorSet.size()];
		final int[] size = new int[colors.length];

		for (int i = 0; i < colors.length; i++)
		{
			int color = colorIter.next();
			if ( color == chosenColor )
			{
				color = colorIter.next();
			}
			colors[i] = color;
			size[i] = colorClass.containsKey(color) ? colorClass.get(color).size() : Integer.MIN_VALUE;
		}

		final int[] indicePermutation = ArrayS.sort(size);
		final int[] sortedColors = new int[colors.length];
		for (int i = 0; i < colors.length; i++)
		{
			sortedColors[i] = colors[indicePermutation[i]];
		}
		return sortedColors;
	}

	public V getBestVertexForKernel()
	{
		V vertex = null;
		int minDe = Integer.MAX_VALUE;
		for (final V v : this.graph.vertexSet())
		{
			int t = this.graph.degreeOf(v);
			for (final V n : this.graph.getNeighbors(v))
			{
				t += this.graph.degreeOf(n);
			}
			if ( t < minDe )
			{
				vertex = v;
				minDe = t;
			}
		}
		return vertex;
	}

	public GraphColoring<V, E> getColoring()
	{
		final int maxC = GreedyColorer.getMaxC(this.coloring.getNumberOfColors());
		return this.getColoring(maxC);
	}

	public GraphColoring<V, E> getColoring(final int parC)
	{
		for (int i = 0; i < IncConsColoringSolver.maxIteration; i++)
		{
			this.iteration++;

			final int numberOfColors = this.coloring.getNumberOfColors();

			final Collection<V> vertices = this.coloring.getSmallestColorClass();
			final int color = this.coloring.getColor(vertices.iterator().next());

			final Collection<V> corruptVertices = new Vector<V>();

			final HashSet<V> maskedVertices = new HashSet<V>();
			maskedVertices.addAll(vertices);

			final int maxC = parC;
			// System.out.println(iteration + "_" + this.coloring.getNumberOfColors() + "(" +
			// vertices + ")" + maxC);
			for (final V vertex : vertices)
			{
				maskedVertices.remove(vertex);

				this.alreadyModifiedVertices.add(vertex);
				final int c = maxC;
				if ( !this.incrementalColoring(this.graph, maskedVertices, vertex, c, color) )
				{
					this.coloring.setColor(vertex, color);
					corruptVertices.add(vertex);
				}
				this.alreadyModifiedVertices.remove(vertex);
			}

			if ( this.coloring.getNumberOfColors() < numberOfColors )
			{
				i = 0;
			}
			else
			{
				for (int f = 0; f < IncConsColoringSolver.maxIteration; f++)
				{
					this.coloring = new GreedyColorer<V, E>(this.graph, new ExtendedGreedy<V, E>(this.coloring,
							GreedyStrategy.RANDOM_SEQUENCE_GREEDY_STRATEGY, IteratedGreedyColorer.RANDOM_PERMUTATION))
							.getColoring();
					this.coloring = new GreedyColorer<V, E>(this.graph, new ExtendedGreedy<V, E>(this.coloring,
							GreedyStrategy.LARGEST_FIRST_GREEDY_STRATEGY, IteratedGreedyColorer.LARGEST_FIRST_PERMUTATION))
							.getColoring();
					if ( !this.coloring.getSmallestColorClass().containsAll(corruptVertices) )
					{
						break;
					}
				}

				if ( this.coloring.getSmallestColorClass().containsAll(corruptVertices) )
				{
					break;
				}

			}
		}
		return this.coloring;
	}

	public UndirectedGraph<VCounter<V>, Edge<VCounter<V>>> getKernel(final int c)
	{
		final V startVertex = this.coloring.getSmallestColorClass().iterator().next();

		final Vector<VCounter<V>> vertexSet = new Vector<VCounter<V>>();
		for (final V v : this.graph.vertexSet())
		{
			if ( v.equals(startVertex) )
			{
				vertexSet.add(new VCounter<V>(v, 0));
			}
			else
			{
				vertexSet.add(new VCounter<V>(v, Integer.MAX_VALUE));
			}
		}

		final Set<VCounter<V>> kernelVertexSet = new TreeSet<VCounter<V>>();

		VCounter<V> vertex = Collections.min(vertexSet);
		while (vertex != null && vertex.getWeight() <= c)
		{
			kernelVertexSet.add(vertex);
			vertexSet.remove(vertex);
			System.out.println(vertex + "(" + vertex.getWeight() + ")");

			{// update weights
				final TreeMap<Integer, Vector<V>> neighboursPerColors = this.graph.getNeighborsPerColor(vertex.vertex,
						this.coloring);
				for (final int color : neighboursPerColors.keySet())
				{
					final Vector<V> neighboursColor = neighboursPerColors.get(color);
					final int newWeight = vertex.getWeight() + neighboursColor.size();
					for (final V n : neighboursColor)
					{
						for (final VCounter<V> v : vertexSet)
						{
							if ( v.vertex.equals(n) )
							{
								if ( newWeight < v.getWeight() )
								{
									System.out.println("set " + v + "(" + newWeight + "<" + v.getWeight() + ")");
								}
								v.setWeight(Math.min(v.getWeight(), newWeight));
							}
						}
					}
				}
			}

			if ( vertexSet.isEmpty() )
			{
				vertex = null;
			}
			else
			{
				vertex = Collections.min(vertexSet);
			}
		}

		final UndirectedGraph<VCounter<V>, Edge<VCounter<V>>> kernelGraph = new ExtendedSimpleGraphAjazenzList<VCounter<V>>(
				kernelVertexSet.size());
		{
			for (final VCounter<V> v : kernelVertexSet)
			{
				kernelGraph.addVertex(v);
				for (final VCounter<V> graphVertex : kernelGraph.vertexSet())
				{
					if ( this.graph.getEdge(v.vertex, graphVertex.vertex) != null )
					{
						kernelGraph.addEdge(v, graphVertex);
					}
				}

			}
		}
		return kernelGraph;
	}

	public GraphColoring<V, E> getUnlieColoring()
	{
		return this.coloring;
	}

	public boolean incrementalColoring(ExtendedUndirectedGraph<V, E> graph, HashSet<V> maskedVertices, V vertex, int c,
			int chosenColor)
	{
		final TreeMap<Integer, Vector<V>> colorClass = IncConsColoringSolver.getNeighborsForVertexPerColor(graph, vertex,
				maskedVertices, this.coloring);
		for (final int i : this.getBestBranchOrder(chosenColor, colorClass))
		{
			this.coloring.setColor(vertex, i);

			if ( !colorClass.containsKey(i) )
				return true;

			final Vector<V> conflicts = new Vector<V>();
			conflicts.addAll(colorClass.get(i));
			if ( this.solve(graph, maskedVertices, conflicts, c, chosenColor) )
				return true;
		}
		return false;
	}

	private boolean solve(ExtendedUndirectedGraph<V, E> graph, HashSet<V> maskedVertices, Vector<V> conflicts, int param,
			int chosenColor)
	{
		if ( conflicts.size() == 0 )
			return true;
		else if ( conflicts.size() > param )
			return false;

		final V firstElement = conflicts.firstElement();
		this.alreadyModifiedVertices.add(firstElement);

		final TreeMap<Integer, Vector<V>> colorClass = IncConsColoringSolver.getNeighborsForVertexPerColor(graph, firstElement,
				maskedVertices, this.coloring);
		final int color = this.coloring.getColor(firstElement);

		for (final int i : this.getBestBranchOrder(chosenColor, colorClass))
		{
			if ( i != color && (!colorClass.containsKey(i) || this.containsNoModifiedElement(colorClass.get(i))) )
			{
				this.coloring.setColor(firstElement, i);
				final Vector<V> conflictTmp = new Vector<V>();
				conflictTmp.addAll(conflicts);
				conflictTmp.remove(firstElement);
				if ( colorClass.containsKey(i) )
				{
					conflictTmp.addAll(colorClass.get(i));
				}

				if ( this.solve(graph, maskedVertices, conflictTmp, param - 1, chosenColor) )
					return true;
			}
		}
		this.alreadyModifiedVertices.remove(firstElement);
		this.coloring.setColor(firstElement, color);
		return false;
	}
}