package dataStructures.graph;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.jgrapht.UndirectedGraph;

public class GraphColoring<V, E>
{
	class ColorClassSize implements Comparator<Collection<V>>
	{
		public int compare(Collection<V> o1, Collection<V> o2)
		{
			return o2.size() - o1.size();
		}

	}

	private static final long serialVersionUID = 1L;

	private final HashMap<Integer, HashSet<V>> colorVertexMap;

	private final UndirectedGraph<V, E> g;

	private final HashMap<V, Integer> vertexColorMap;

	public GraphColoring(final UndirectedGraph<V, E> g)
	{
		this.g = g;
		this.vertexColorMap = new HashMap<V, Integer>(g.vertexSet().size());
		this.colorVertexMap = new HashMap<Integer, HashSet<V>>();

	}

	public int getColor(V v)
	{
		return this.vertexColorMap.get(v);
	}

	public HashMap<Integer, HashSet<V>> getColorVertexMapping()
	{
		return this.colorVertexMap;
	}

	public int getNumberOfColors()
	{
		return this.getColorVertexMapping().keySet().size();
	}

	public Collection<V> getSmallestColorClass()
	{
		final HashMap<Integer, HashSet<V>> map = this.getColorVertexMapping();
		return (Collection<V>) Collections.max(map.values(), new ColorClassSize()).clone();

	}

	public boolean isAlreadyColored(V v)
	{
		return this.vertexColorMap.containsKey(v);
	}

	/**
	 * Checks whether the coloring, represented by this class, is a proper vertex-coloring or not.
	 * 
	 * @return
	 */
	public boolean isCorrectColoring(boolean quiet)
	{
		boolean flag = true;
		for (final E e : this.g.edgeSet())
		{
			final V edgeSource = this.g.getEdgeSource(e);
			final V edgeTarget = this.g.getEdgeTarget(e);
			if ( !this.vertexColorMap.containsKey(edgeSource) || !this.vertexColorMap.containsKey(edgeTarget) )
			{
				flag = false;
				if ( !quiet )
				{
					System.out.print(!this.vertexColorMap.containsKey(edgeSource) ? edgeSource : edgeTarget + "|");
				}
			}
			else
			{
				final int color1 = this.getColor(edgeSource);
				final int color2 = this.getColor(edgeTarget);
				if ( color1 == color2 )
				{
					flag = false;
					if ( !quiet )
					{
						System.out.println("conflict:" + edgeSource + "(" + color1 + ")_" + edgeTarget + "(" + color2 + ")");
					}
				}
			}
			if ( quiet && !flag )
			{
				break;
			}
		}
		return flag;
	};

	public boolean isCorrectColoring(boolean quiet, Set<V> maskedVertices)
	{
		boolean flag = true;
		for (final E e : this.g.edgeSet())
		{
			if ( !maskedVertices.contains(this.g.getEdgeSource(e)) && !maskedVertices.contains(this.g.getEdgeTarget(e)) )
			{
				final int color1 = this.getColor(this.g.getEdgeSource(e));
				final int color2 = this.getColor(this.g.getEdgeTarget(e));
				if ( color1 == color2 )
				{
					flag = false;
					if ( !quiet )
					{
						System.out.println("conflict:" + this.g.getEdgeSource(e) + "(" + color1 + ")_" + this.g.getEdgeTarget(e)
								+ "(" + color2 + ")");
					}
					else
					{
						break;
					}

				}
			}
		}
		return flag;
	}

	public void removeColor(V v)
	{
		if ( this.colorVertexMap.containsKey(this.getColor(v)) )
		{
			final HashSet<V> colorclass = this.colorVertexMap.get(this.getColor(v));
			colorclass.remove(v);
			if ( colorclass.isEmpty() )
			{
				this.colorVertexMap.remove(this.getColor(v));
			}
		}

		this.vertexColorMap.remove(v);
	}

	public void setColor(V v, int color)
	{
		if ( this.vertexColorMap.containsKey(v) )
		{
			final int oldColor = this.vertexColorMap.get(v);

			this.colorVertexMap.get(oldColor).remove(v);
			if ( this.colorVertexMap.get(oldColor).isEmpty() )
			{
				this.colorVertexMap.remove(oldColor);
			}
		}

		this.vertexColorMap.put(v, color);
		HashSet<V> colorClass = this.colorVertexMap.get(color);
		if ( colorClass == null )
		{
			colorClass = new HashSet<V>();
			this.colorVertexMap.put(color, colorClass);
		}
		colorClass.add(v);
	}

	@Override
	public String toString()
	{
		return "c:" + this.getNumberOfColors() + ";" + this.getColorVertexMapping().toString();
	}
}