package distanceFunction;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Vector;

import distanceFunction.random.RandomDistanceFunctions;

/**
 * Representation of a distance function.
 * 
 * @author Sepp Hartung
 * @version 1.0
 */
public class SymmetricDiscreteDistanceFunction
{

	public final static String sep = ",";

	public static int[][] copyArray(int[][] arr)
	{
		final int[][] distance = new int[arr.length][];
		for (int i = 0; i < arr.length; i++)
		{
			distance[i] = arr[i].clone();
		}
		return distance;
	}

	public static SymmetricDiscreteDistanceFunction createInstanceFromFile(String filePath) throws IOException
	{
		final LineNumberReader reader = new LineNumberReader(new FileReader(new File(filePath)));
		String line;
		final Vector<int[]> results = new Vector<int[]>();
		while ((line = reader.readLine()) != null)
		{
			// ignore first (description) line
			if ( reader.getLineNumber() > 1 && line.length() != 0 )
			{
				final String[] arrS = line.split(SymmetricDiscreteDistanceFunction.sep);
				final int[] arr = new int[arrS.length];
				for (int i = 0; i < arrS.length; i++)
				{
					arr[i] = Integer.parseInt(arrS[i]);
				}
				results.add(arr);
			}
		}
		reader.close();
		return new SymmetricDiscreteDistanceFunction(results.toArray(new int[0][0]));
	}

	public static void visualizeUltrametric(String filePath, SymmetricDiscreteDistanceFunction distanceFunction)
			throws IOException
	{
		class Cluster
		{

			int index;

			protected String name;

			public Cluster(String i, int index)
			{
				this.name = i;
				this.index = index;
			}
		}
		;
		class SuperCluster extends Cluster
		{

			private final Vector<Cluster> clusters = new Vector<Cluster>();

			public SuperCluster(String name, Cluster clu)
			{
				super(name, clu.index);
				this.clusters.add(clu);
			}

			public void add(Cluster nextCluster)
			{
				this.clusters.add(nextCluster);
			}

			public void writeToFile(FileWriter writer, int distance) throws IOException
			{
				for (final Cluster i : this.clusters)
				{
					writer.write(this.name + " -- " + i.name + "\n");
				}
			}
		}
		;

		Vector<Cluster> clusters = new Vector<Cluster>();
		for (int i = 0; i < distanceFunction.getNoOfElements(); i++)
		{
			final Cluster clu = new Cluster(String.valueOf(i), i);
			clusters.add(clu);
		}
		final FileWriter writer = new FileWriter(new File(filePath));
		writer.write("graph G {\n");
		writer.write("ranksep = 2\n");

		int distance = 1;
		int count = 0;
		while (clusters.size() > 1)
		{
			final Vector<Cluster> newClusters = new Vector<Cluster>();
			while (!clusters.isEmpty())
			{
				final SuperCluster cluster = new SuperCluster("n" + count++, clusters.remove(0));

				final Iterator<Cluster> iter = clusters.iterator();
				while (iter.hasNext())
				{
					final Cluster nextCluster = iter.next();
					if ( distanceFunction.get(cluster.index, nextCluster.index) == distance )
					{
						cluster.add(nextCluster);
						iter.remove();
					}
				}
				cluster.writeToFile(writer, distance);
				writer.flush();
				newClusters.add(cluster);
			}
			clusters = newClusters;
			distance++;
		}
		writer.write("}");
		writer.close();

	}

	private final int[][] distanceFunction;

	private final boolean[] isDeletedData;

	private final int maxVal;

	private final int noElements;

	public SymmetricDiscreteDistanceFunction(int n)
	{
		this(n, 0, -1);
	}

	/**
	 * Constructs a new distance function with randomized values.
	 * 
	 * @param n number of nodes
	 * @param maxDistValue range [1,...,maxDistValue]
	 */
	public SymmetricDiscreteDistanceFunction(int n, int maxDistValue, int variation)
	{
		this(RandomDistanceFunctions.getRandomDistanceFunction(n, maxDistValue, variation,
				RandomDistanceFunctions.DEFAULT_MAX_DEGREE));
	}

	public SymmetricDiscreteDistanceFunction(int[] nodes, SymmetricDiscreteDistanceFunction dist)
	{
		this(nodes.length, 0, -1);
		for (int i = 0; i < nodes.length; i++)
		{
			for (int j = i + 1; j < nodes.length; j++)
			{
				this.set(i, j, dist.get(nodes[i], nodes[j]));
			}
		}
	}

	/**
	 * Constructs a new distance function.
	 * 
	 * @param distanceFunction
	 */
	public SymmetricDiscreteDistanceFunction(int[][] distanceFunction)
	{
		this.distanceFunction = distanceFunction;
		this.isDeletedData = new boolean[this.distanceFunction.length + 1];
		Arrays.fill(this.isDeletedData, false);
		this.noElements = this.distanceFunction.length + 1;
		// traverseArray and Set Max Value
		int mv = 0;

		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				if ( this.get(i, j) > mv )
				{
					mv = this.get(i, j);
				}
			}
		}
		this.maxVal = mv;
	}

	/**
	 * Copy constructor
	 * 
	 * @param function
	 */
	public SymmetricDiscreteDistanceFunction(SymmetricDiscreteDistanceFunction function)
	{
		this(SymmetricDiscreteDistanceFunction.copyArray(function.distanceFunction));
	}

	public void deleteData(int index)
	{
		this.setDeleteStatus(index, true);
	}

	@Override
	public boolean equals(Object obj)
	{
		if ( this == obj )
			return true;
		if ( obj == null )
			return false;
		if ( this.getClass() != obj.getClass() )
			return false;
		final SymmetricDiscreteDistanceFunction other = (SymmetricDiscreteDistanceFunction) obj;
		if ( !Arrays.equals(this.distanceFunction, other.distanceFunction) )
			return false;
		return true;
	}

	/**
	 * Returns d(i,j).
	 * 
	 * @param i index
	 * @param j index
	 * @return
	 */
	public int get(int i, int j)
	{
		if ( j < i )
			return this.distanceFunction[j][i - (j + 1)];
		else
			return this.distanceFunction[i][j - (i + 1)];
	}

	/**
	 * Computes the similarity of two distance functions under the l1-norm.
	 */
	public int getDistance(SymmetricDiscreteDistanceFunction dis)
	{
		int dist = 0;
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				dist += Math.abs(this.get(i, j) - dis.get(i, j));
			}
		}
		return dist;
	}

	public int getMaxValue()
	{
		return this.maxVal;
	}

	public int getNoOfElements()
	{
		return this.noElements;
	}

	@Override
	public int hashCode()
	{
		final int PRIME = 31;

		int hash = 0;
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{

				hash += PRIME * this.get(i, j) * (i + PRIME) * (j + PRIME);
			}
		}
		return hash;
	}

	/**
	 * @param i
	 * @param j
	 * @param k
	 */
	public void increase(int i, int j, int k)
	{
		this.set(i, j, this.get(i, j) + k);

	}

	/**
	 * Check whether the distances between the given indices form a conflict or not.
	 * 
	 * @param indices, such that d_indices[0]_indices[1]=max
	 * @return
	 */
	public int[] isConflict(int i, int j, int l)
	{
		final int d_i_j = this.get(i, j);
		final int d_i_l = this.get(i, l);
		final int d_j_l = this.get(j, l);
		if ( d_i_j > d_i_l && d_i_j > d_j_l )
			return new int[] { Math.min(i, j), Math.max(i, j), l };
		else if ( d_i_l > d_i_j && d_i_l > d_j_l )
			return new int[] { Math.min(i, l), Math.max(i, l), j };
		else if ( d_j_l > d_i_j && d_j_l > d_i_l )
			return new int[] { Math.min(j, l), Math.max(j, l), i };
		return null;
	}

	public boolean isConflict(int[] indices)
	{
		final int[] arr = this.isConflict(indices[0], indices[1], indices[2]);
		if ( arr != null )
		{
			indices[0] = arr[0];
			indices[1] = arr[1];
			indices[2] = arr[2];
			return true;
		}
		return false;
	}

	public boolean isDeletedData(int i)
	{
		return this.isDeletedData[i];
	}

	public boolean isUltrametric()
	{
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				for (int t = j + 1; t < this.getNoOfElements(); t++)
				{
					if ( this.isConflict(i, j, t) != null )
						return false;
				}
			}
		}
		return true;
	}

	public boolean isUltrametric(Vector<Integer> indices)
	{
		for (int i = 0; i < indices.size(); ++i)
		{
			for (int j = i + 1; j < indices.size(); ++j)
			{
				for (int k = j + 1; k < indices.size(); ++k)
				{
					if ( this.isConflict(indices.get(i), indices.get(j), indices.get(k)) != null )
						return false;
				}
			}
		}
		return true;
	}

	/**
	 * @param i
	 * @param j
	 * @param k
	 */
	public void set(int i, int j, int k)
	{
		if ( j < i )
		{
			this.distanceFunction[j][i - (j + 1)] = k;
		}
		else
		{
			this.distanceFunction[i][j - (i + 1)] = k;
		}
	}

	public void setDeleteStatus(int index, boolean del)
	{
		this.isDeletedData[index] = del;
	}

	@Override
	public String toString()
	{
		final StringBuffer s = new StringBuffer();
		for (int i = 0; i <= this.distanceFunction.length; i++)
		{
			s.append(i + "\t");
		}
		s.append('\n');
		for (int i = 0; i <= this.distanceFunction.length; i++)
		{
			s.append("_______");
		}
		s.append('\n');
		for (int i = 0; i < this.distanceFunction.length; i++)
		{
			final int[] arr = this.distanceFunction[i];
			s.append(i + "|\t");
			for (int t = 0; t < i; t++)
			{
				s.append("\t");
			}
			for (int j = 0; j < arr.length; j++)
			{
				if ( j == 0 )
				{
					s.append("{");
				}
				s.append(arr[j]);

				if ( j < arr.length - 1 )
				{
					s.append(",");
					s.append("\t");
				}
				else
				{
					s.append("},");
				}
			}
			s.append('\n');
		}
		s.append('\n');
		return s.toString();
	}

	public void writeToFile(String filePath) throws IOException
	{
		this.writeToFile(filePath, null);
	}

	public void writeToFile(String filePath, String[] par) throws IOException
	{
		final FileWriter writer = new FileWriter(new File(filePath));
		// description
		writer.write("# n=" + this.getNoOfElements());
		if ( par != null )
		{
			for (int i = 0; i < par.length && i + 1 < par.length; i += 2)
			{
				writer.write("\t");
				writer.write(par[i] + "=" + par[i + 1]);
			}
		}
		for (final int[] element : this.distanceFunction)
		{
			writer.write("\n");
			for (int j = 0; j < element.length; j++)
			{
				if ( j != 0 )
				{
					writer.write(SymmetricDiscreteDistanceFunction.sep);
				}
				writer.write(String.valueOf(element[j]));
			}
		}
		writer.close();
	}
}